//----------------------------------------------------------------------------
//	Copyright (C) 2002 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file ExecutionList.cpp

	\author Ralf Gerstenberger
	<!-- [\author <author>]* -->

	\date created at 2002/01/25

	\brief Implementation of ExecutionList

	\sa ExecutionList.h

	<!-- [detailed description] -->

	<!-- [\todo {todos for this file}]* -->

	\since 1.0
*/

#include <odemx/base/Process.h>
#include <odemx/base/ExecutionList.h>
#include <odemx/base/InSort.h>
#include <odemx/util/ErrorHandling.h>

#include <algorithm>

using namespace std;
using namespace odemx;

ExecutionList::ExecutionList(ExecutionListObserver* o) : Observable<ExecutionListObserver>(o) {
	_obsForEach(ExecutionListObserver, Create(this));
}

ExecutionList::~ExecutionList() {
	_obsForEach(ExecutionListObserver, Destroy(this));
}

void ExecutionList::addProcess(Process* p) {
	// attach Process p at SimTime to ExecutionList (FIFO), considering priority

	if (p==0) {
		// error: illegal parameter
		error("addProcess(); invalid argument");
		return;
	}

	// adjust execution time
	if (p->getExecutionTime() < getTime())
		p->setExecutionTime(getTime());

	inSort(p);

	_obsForEach(ExecutionListObserver, AddProcess(this, p));
}

void ExecutionList::insertProcess(Process* p) {
	// insert Process p at SimTime in ExecutionList (LIFO), considering priority

	if (p==0) {
		// error: illegal parameter
		error("insertProcess(); invalid argument");
		return;
	}

	// adjust execution time
	if (p->getExecutionTime() < getTime())
		p->setExecutionTime(getTime());

	inSort(p, false);
	_obsForEach(ExecutionListObserver, InsertProcess(this, p));	
}

void ExecutionList::insertProcessAfter(Process* p, Process* previous) {
	// insert Process p after 'previous', changing priority if necessary

	std::list<Process*>::iterator pre;
	std::list<Process*>::iterator i;
	
	if (p==0) {
		// error: no Process to enter
		error("insertProcessAfter(); invalid argument: Process* p");
		return;
	}

	if (previous==0) {
		// error: no 'previous'
		error("insertProcessAfter(); invalid argument: Process* previous");
		return;
	}

	if (p==previous)
		return;

	// find 'previous'
	pre = find(l.begin(), l.end(), previous);
	if (pre==l.end()) {
		// error: 'previous' is not listed
		error("insertProcessAfter(); 'previous' not listed");
		return;
	}

	// remove Process p from ExecutionList
	i = find(l.begin(), l.end(), p);
	if (i!=l.end())
		l.erase(i);

	// adjust priority
	if (p->getPriority()>(*pre)->getPriority())
		p->setPriority((*pre)->getPriority());

	// set execution time
	p->setExecutionTime((*pre)->getExecutionTime());

	// insert Process p after 'pre' (before 'next' of 'pre')
	++pre;
	l.insert(pre, p);

	_obsForEach(ExecutionListObserver, InsertProcessAfter(this, p, previous));
}

void ExecutionList::insertProcessBefore(Process* p, Process* next) {
	// insert Process p before 'next', changing priority if necessary

	std::list<Process*>::iterator nex;
	std::list<Process*>::iterator i;
	
	if (p==0) {
		// error: no Process to enter
		error("insertProcessBefore(); invalid argument: Process* p");
		return;
	}

	if (next==0) {
		// error: no 'next'
		error("insertProcessBefore(); invalid argument: Process* next");
		return;
	}

	if (p==next)
		return;

	// find 'next'
	nex = find(l.begin(), l.end(), next);
	if (nex==l.end()) {
		// error: 'next' is not listed
		error("insertProcessBefore(); 'next' not listed");
		return;
	}

	// remove Process p from ExecutionList
	i = find(l.begin(), l.end(), p);
	if (i!=l.end())
		l.erase(i);

	// adjust priority
	if (p->getPriority()<(*nex)->getPriority())
		p->setPriority((*nex)->getPriority());

	// set execution time
	p->setExecutionTime((*nex)->getExecutionTime());

	// insert Process p before 'next'
	l.insert(nex, p);

	_obsForEach(ExecutionListObserver, InsertProcessBefore(this, p, next));
}

void ExecutionList::removeProcess(Process* p) {
	l.remove(p);

	_obsForEach(ExecutionListObserver, RemoveProcess(this, p));
}

Process* ExecutionList::getNextProcess() {
	if(isEmpty())
		return 0;
	
	return l.front();
}

void ExecutionList::inSort(Process* e, bool fifo/* = true*/) {
	InSort(l, e, fifo);
}

