//----------------------------------------------------------------------------
//	Copyright (C) 2002 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/** \file ExecutionList.h	

	\author Ralf Gerstenberger
	<!-- [\author <author>]* -->

	\date created at 2002/01/23

	\brief Declaration of ExecutionList and ExecutionListObserver

	\sa ExecutionList.cpp

	<!-- [detailed description] -->

	<!-- [\todo {todos for this file}]* -->

	\since 1.0
*/

#ifndef ODEMX_EXECUTIONLIST_INCLUDED
#define ODEMX_EXECUTIONLIST_INCLUDED

#include <odemx/util/SimTime.h>
#include <odemx/util/Observable.h>

#include <list>

namespace odemx {
	class Process;
	class ExecutionListObserver;

	/** \class ExecutionList

		\ingroup base

		\author Ralf Gerstenberger
		<!-- [\author <author>]* -->

		\brief ExecutionList implements an execution schedule for process objects.

		\note ExecutionList supports Observation.

		<!-- [\sa {references to other classes}]* -->

		ExecutionList is used to manage the process execution order. A process
		is scheduled at a given time considering its priority and a FIFO- or LIFO-
		strategy, or in relation to an already scheduled process.

		<!-- [\warning {warnings}]* -->

		<!-- [\todo {todos for this file}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class ExecutionList : public Observable<ExecutionListObserver> {
	public:
		/// Construction
		ExecutionList(ExecutionListObserver* o = 0);
		/// Destruction
		~ExecutionList();

		/**
			\name Process Management

			@{
		*/
		/**
			\brief append Process 

			\param p
				pointer to Process
			
			Appends Process \p p at its execution time.
			If there are already Process objects scheduled at 
			the execution time of \p p, addProcess will consider 
			the priorities and use the FIFO strategy.

			\sa insertProcess
		*/
		void addProcess(Process* p);

		/**
			\brief insert Process

			\param p
				pointer to Process

			Inserts Process \p p at its execution time.
			If there are already Process objects scheduled at 
			the execution time of \p p, insertProcess will consider 
			the priorities and use the LIFO strategy.

			\sa addProcess
		*/
		void insertProcess(Process* p);

		/**
			\brief insert Process \p p after \p previous

			\param p
				pointer to new Process

			\param previous
				pointer to Process after which the new Process 
				should be inserted

			Inserts Process \p p after the \p previous Process
			in ExecutionList.

			\note insertProcessAfter will change the priority of
			Process \p p if necessary.

			\sa insertProcessBefore
		*/
		void insertProcessAfter(Process* p, Process* previous);

		/**
			\brief insert Process \p p before \p next

			\param p
				pointer to new Process

			\param next
				pointer to Process before which the new Process 
				should be inserted
			
			Inserts Process \p p before the \p next Process
			in ExecutionList.

			\note insertProcessBefore will change the priority of
			Process \p p if necessary.

			\sa insertProcessAfter
		*/
		void insertProcessBefore(Process* p, Process* next);

		/**
			\brief remove Process

			Removes Process \p p from ExecutionList.
		*/
		void removeProcess(Process* p);
		//@}

		/**
			\brief to most Process in ExecutionList

			\return
				pointer to top most Process in ExecutionList
		*/
		Process* getNextProcess();

		/**
			\brief check if ExecutionList is empty
		*/
		bool isEmpty() {return l.empty();}

		/**
			\brief get simulation time

			\return
				current simulation time
		*/
		virtual SimTime getTime() const = 0;

		// Implementation
	private:
		/// Process list
		std::list<Process*> l;

		friend class Process;

		/**
			\brief inSort Process \p e in Process list

			\param e
				pointer to Process

			\param fifo
				use FIFO or LIFO strategy for inSort
		*/
		void inSort(Process* e, bool fifo = true);
	};

	/** \interface ExecutionListObserver

		\author Ralf Gerstenberger
		<!-- [\author <author>]* -->

		\brief Observer for ExecutionList events.

		<!-- [\note {notes}]* -->

		\sa ExecutionList

		<!-- [detailed description] -->

		<!-- [\warning {warnings}]* -->

		<!-- [\todo {todos for this file}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class ExecutionListObserver {
	public:
		/// construction
		virtual void onCreate(ExecutionList* sender) {}
		/// destruction
		virtual void onDestroy(ExecutionList* sender) {}
		/// Process added
		virtual void onAddProcess(ExecutionList* sender, Process* newProcess) {}
		/// Process insertion
		virtual void onInsertProcess(ExecutionList* sender, Process* newProcess) {}
		/// Process insertion with parameters
		virtual void onInsertProcessAfter(ExecutionList* sender, Process* newProcess, Process* previousProcess) {}
		/// Process insertion with parameters
		virtual void onInsertProcessBefore(ExecutionList* sender, Process* newProcess, Process* nextProcess) {}
		/// Process removed
		virtual void onRemoveProcess(ExecutionList* sender, Process* removedProcess) {}
	};
}

#endif

